/*!\file lraam_generalization.c
 * 
 * \author Michal Pokorny
 * \date 02/2006 - 03/2006
 *
 * \brief Utility supporting generalization procedure
 *
 * Given experiment status on standard input (with weight matrix evolved to some extent),
 * generalization procedure is run with bounding parameters specified within option.
 * First of all, user is encouraged to specify theta - reconstruction treshold.
 * Moreover, user can specify maximal depth and maximum number of leaves (length)
 * for considered new structures. User can also specify the format for output
 * of all encodable structures.
 */

#include <getopt.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lraam.h"

#define MAX_PATH_LENGTH 128
#define MAX_STRUCT	8192

#define DEFAULT_THETA	0.7
#define DEFAULT_DEPTH	-1
#define DEFAULT_LENGTH 	-1
#define DEFAULT_VERBOSE	0
#define DEFAULT_PRINT_ALL 0

lraam_t * raam;
set_t * dataset;
set_t * termset;
set_t * genset;

// command line options defaults
// -----------------------------------------------------------------------------
value_t theta = DEFAULT_THETA;		/* switch --theta, 	-t	*/
int depth = DEFAULT_DEPTH;		/* switch --depth, 	-d	*/
int length = DEFAULT_LENGTH;		/* switch --length, 	-l	*/
bool_t verbose = DEFAULT_VERBOSE;	/* switch --verbose, 	-v	*/
bool_t print_all = DEFAULT_PRINT_ALL;	/* switch --printall, 	-p	*/

// prints usage of this program
// -----------------------------------------------------------------------------
void
print_usage () {

    printf ("Usage:\n");
    printf ("  lraam_generalization [OPTIONS]\n");
    printf ("\nOptions:\n");
    printf ("  -t <theta>, --theta=<theta>\n");
    printf ("       Set theta parameter.\n");
    printf ("  -d <depth>, --depth=<depth>\n");
    printf ("       Set maximum depth of generated structures.\n");
    printf ("  -l <length>, --length=<length>\n");
    printf ("       Set maximum length (number of leaves - terminal symbols)\n");
    printf ("       of generated structures.\n");
    printf ("  -v, --verbose\n");
    printf ("       Set this flag to get temporary information about\n");
    printf ("       progress of generalization.\n");
    printf ("  -p, --printall\n");
    printf ("       Set this flag to force printing all structures\n");
    printf ("       that the network can represent.\n");
    printf ("  -s, --settings\n");
    printf ("       Display default option values and exit.\n");
    printf ("  -h, --help\n");
    printf ("       Display this help and exit.\n");
    
    return;
}


// release memory allocated when loading experiment
// -----------------------------------------------------------------------------
void free_all() {

    lraam_free(raam);
    set_free(dataset);
    set_free(termset);
    set_free(genset);

    return;
}


// prints default values of command line options
// -----------------------------------------------------------------------------
void
print_defaults () {

    printf ("Default values of command line options:\n");

    // theta parameter
    printf ("  <theta> = ");
    value_fprintf(stdout, theta);
    fprintf(stdout, "\n");
    // maximal depth
    printf ("  <depth> = ");
    if (depth == -1) printf("no bounds\n");
    else printf("%d\n", depth);
    // maximal length
    printf ("  <length> = ");
    if (length == -1) printf("no bounds\n");
    else printf("%d\n", length);
    // verbose mode
    printf ("  <verbose> = %s\n", verbose?"yes":"no");
    // print-all mode
    printf ("  <print_all> = %s\n", print_all?"yes":"no");
    
    return;
}


// parses command line options
// -----------------------------------------------------------------------------
void
parse_options(int argc, char **argv) {
    int c;
    static const struct option long_options[] = {
	{"theta", required_argument, NULL, 't'},
	{"depth", required_argument, NULL, 'd'},
	{"length", required_argument, NULL, 'l'},
	{"verbose", no_argument, NULL, 'v'},
	{"printall", no_argument, NULL, 'p'},
	{"settings", no_argument, NULL, 's'},
	{"help", no_argument, NULL, 'h'},
	{0, 0, 0, 0}
    };

    static const char short_options[] = "t:d:l:vpsh";

    while(1) {
	c = getopt_long(argc, argv, short_options, long_options, NULL);
	if (c == -1) break;
	switch (c) {
	    case 't':
		sscanf(optarg, "%lf", &theta);
		break;
	    case 'd':
		sscanf(optarg, "%d", &depth);
		break;
	    case 'l':
		sscanf(optarg, "%d", &length);
		break;
	    case 'v':
		verbose = 1;
		break;
	    case 'p':
		print_all = 1;
		break;
	    case 's':
		print_defaults();
		exit(0);
		break;
	    case 'h':
	    case '?':
	    default:
		print_usage();
		exit(1);
	}
    }
    
    return;
}


// main
// -----------------------------------------------------------------------------
int
main(int argc, char **argv) {

    error_t err;
    
    // get command line options
    parse_options(argc, argv);

    // load experiment
    err = experiment_load(stdin, &raam, &dataset, &termset);
    if (err != ERR_OK) {
	fprintf(stderr, "ERROR %d: cannot load experiment.\n\n", err);
	exit(err);
    }

    // set theta - reconstruction treshold
    lraam_set_theta(raam, theta);
    
    // allocate some memory
    genset = set_alloc(MAX_STRUCT, STRUCTURE);
    if (genset == NULL) {
	fprintf(stderr, "ERROR: genset: not enough memory.\n\n");
	exit(1);
    }

    // generalization
    lraam_generalize(raam, genset, dataset, termset, length, depth, verbose);
    
    // do some output
    if (print_all) {
	// in this case, terminal symbols and (possibly) structures from dataset
	// will be on the list
	set_save_structures(stdout, genset, termset, 0);
    } else {
	// don't account terminals as regular structures
	// possibly structures from dataset will be within genset
	fprintf(stdout, "%d\n", genset->no_of_objects - termset->no_of_objects);
    }

    // release memory
    free_all();

    return 0;
}
