/*! \file lraam_init.c
 *
 *  \author Michal Pokorny
 *  \date 02/2006 - 03/2006
 *
 *  \brief Utility for generating initial conditions for experiment
 *         with <em>linear RAAM</em> (or <em>PCA-RAAM</em>).
 *
 *  The concept of \b experiment is as follows: when one provides simulation
 *  with network, in most cases one uses the same training set (or \b dataset)
 *  and the same encoding of terminal symbols (or \b termset) throughout 
 *  the learning process. So termset and dataset are stored. \n
 *  In order to be able to reproduce simulation (or to be able to continue
 *  from any point in different way), it is necesarry to store learning history,
 *  i.e. for each block number of learning iterations and learning rate, and
 *  random generator history, i.e. randseed and number of random calls. \n
 *  It is also useful to store arity of structures and dimension of encoding vectors.
 */

#include <getopt.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lraam.h"

#define MAX_PATH_LENGTH 128

#define DEFAULT_TERMSET 	"termset"
#define DEFAULT_DATASET 	"dataset"
#define DEFAULT_RANDSEED 	-1
#define DEFAULT_PRECISION	-1
#define DEFAULT_WBOUNDS		-1

lraam_t * raam;
set_t * dataset;
set_t * termset;

int arity = -1;
int dimension = -1;

// command line options defaults
// -----------------------------------------------------------------------------
char termset_filename[MAX_PATH_LENGTH + 1] = DEFAULT_TERMSET;	
					/* switch --termset,	-t	*/
char dataset_filename[MAX_PATH_LENGTH + 1] = DEFAULT_DATASET;
					/* switch --dataset,	-d	*/
int randseed = DEFAULT_RANDSEED;	/* switch --randseed, 	-r 	*/
int precision = DEFAULT_PRECISION;	/* switch --precision, 	-p 	*/
double wbounds = DEFAULT_WBOUNDS;	/* switch --weight-bounds,	-w 	*/

// prints usage of this program
// -----------------------------------------------------------------------------
void
print_usage () {

    printf ("Usage:\n");
    printf ("  lraam_init [OPTIONS]\n");
    printf ("\nOptions:\n");
    printf ("  -t <termset_filename>, --termset=<termset_filename>\n");
    printf ("       File with terminals and their encodings.\n");
    printf ("  -d <dataset_filename>, --dataset=<dataset_filename>\n");
    printf ("       File with structures in training set. Each terminal\n");
    printf ("       with occurence in dataset  must have specification\n");
    printf ("       in termset.\n");
    printf ("  -r <randseed>, --randseed=<randseed>\n");
    printf ("       Specify the randseed used for setting up random generator.\n");
    printf ("       Leaving this option off, random generator will be set up\n");
    printf ("       with current timestamp.\n");
    printf ("  -p <precision>, --precision=<precision>\n");
    printf ("       Specify the precision used for printing value_t variables.\n");
    printf ("       You should use this option when you are using high precision\n");
    printf ("       values for terminal encodings.\n");
    printf ("  -w <weight_bounds>, --weight-bounds=<weight_bounds>\n");
    printf ("       All weights will be initialized from interval <-wb; wb>.\n");
    printf ("  -s, --settings\n");
    printf ("       Display default values of options.\n");
    printf ("  -h, --help\n");
    printf ("       Display this help and exit.\n");
    
    return;
}


// prints default values of command line options
// -----------------------------------------------------------------------------
void
print_defaults () {
    FILE * f;

    printf ("Default values of command line options:\n");

    // termset
    printf ("  <termset_filename> = %s", termset_filename);
    f = fopen(termset_filename, "r");
    if (f != NULL) {
	printf("\n");
	fclose(f);
    } else {
	printf("\tERROR: %d\n", errno);
    }
    
    // dataset
    printf ("  <dataset_filename> = %s", dataset_filename);
    f = fopen(dataset_filename, "r");
    if (f != NULL) {
	printf("\n");
	fclose(f);
    } else {
	printf("\tERROR: %d\n", errno);
    }
    
    // randseed
    printf ("  <randseed> = ");
    if (randseed == -1) printf("time(NULL)\n");
    else printf("%d\n", randseed);
    
    // precision
    printf ("  <precision> = ");
    if (precision == -1) printf("default (3)\n");
    else printf("%d\n", precision);
    
    // randseed
    printf ("  <weight_bounds> = ");
    if (wbounds < 0) printf("default (0.5)\n");
    else printf("%.5lf\n", wbounds);
    
    return;
}


// parses command line options
// -----------------------------------------------------------------------------
void
parse_options(int argc, char **argv) {
    int c;
    static const struct option long_options[] = {
	{"termset", required_argument, NULL, 't'},
	{"dataset", required_argument, NULL, 'd'},
	{"randseed", required_argument, NULL, 'r'},
	{"precision", required_argument, NULL, 'p'},
	{"weight-bounds", required_argument, NULL, 'w'},
	{"settings", no_argument, NULL, 's'},
	{"help", no_argument, NULL, 'h'},
	{0, 0, 0, 0}
    };

    static const char short_options[] = "t:d:r:p:w:sh";

    while(1) {
	c = getopt_long(argc, argv, short_options, long_options, NULL);
	if (c == -1) break;
	switch (c) {
	    case 't': 
		strncpy(termset_filename, optarg, MAX_PATH_LENGTH);
		termset_filename[MAX_PATH_LENGTH] = '\0';
		break;
	    case 'd': 
		strncpy(dataset_filename, optarg, MAX_PATH_LENGTH);
		dataset_filename[MAX_PATH_LENGTH] = '\0';
		break;
	    case 'r': 
		sscanf(optarg, "%d", &randseed);
		if (randseed < 0) randseed = DEFAULT_RANDSEED;
		break;
	    case 'p': 
		sscanf(optarg, "%d", &precision);
		if (precision < 0 || precision > 10) precision = DEFAULT_PRECISION;
		break;
	    case 'w': 
		sscanf(optarg, "%lf", &wbounds);
		if (wbounds < 0) wbounds = DEFAULT_WBOUNDS;
		break;
	    case 's':
		print_defaults();
		exit(0);
		break;
	    case 'h':
	    case '?':
	    default:
		print_usage();
		exit(1);
	}
    }
    
    return;
}


// check for validity of params
// -----------------------------------------------------------------------------
void
check() {

    FILE * f;
    int i, n;
    char str[4096];
    char * s, * token;
    int oldd, newd;
    int olda, newa;

    // checking arity
    fprintf(stderr, "checking arity..");
    f = fopen(dataset_filename, "r");
    if (f == NULL) {
        fprintf(stderr, "ERROR: dataset file not opened.\n\n");
        print_usage();
        exit(1);
    }
    fscanf(f, "%d\n", &n);			// number of structures in dataset
    olda = -1;					// olda - arity of structures yet read
    for (i=0; i<n; i++) {
        fscanf(f, "%[^\n]\n", str);		// read structure
        s = str;				// better not to change str
        newa = 0;				// arity of current
        token = strtok(s, " ");
        while (token != NULL) {
    	    if (!strcmp(token, "(")) newa = -1;
    	    if (!strcmp(token, ")")) break;
    	    newa++;
	    token = strtok(NULL, " ");
	}
	if (olda == -1) olda = newa;	// first loop
	if (newa != olda) {
	    fprintf(stderr, "ERROR: cannot determine arity from dataset file \"%s\".\n\n", dataset_filename);
	    fclose(f);
	    exit(1);
	}
    }
    arity = newa;
    fclose(f);
    fprintf(stderr, "done: %d\n", arity);
    
    // checking dimension
    fprintf(stderr, "checking dimension..");
    f = fopen(termset_filename, "r");
    if (f == NULL) {
        fprintf(stderr, "ERROR: termset file not opened.\n\n");
        print_usage();
        exit(1);
    }
    fscanf(f, "%d\n", &n);			// number of structures in dataset
    oldd = -1;				// oldd - dimension of vectors yet read
    for (i=0; i<n; i++) {
        fscanf(f, "%[^\n]\n", str);		// read terminal label with vectors
        s = str;				// better not to change str
        newd = 0;				// dimension of current terminal
        token = strtok(s, " \t");
        while (token != NULL) {
    	    newd++;				// count all tokens
	    token = strtok(NULL, " \t");
	}
	if (oldd == -1) oldd = newd;	// first loop
	if (oldd != newd) {			// if dimension of current terminal differs from previos
	    fprintf(stderr, "ERROR: cannot determine dimension from termset file \"%s\".\n\n", termset_filename);
	    fclose(f);
	    exit(1);
	}
    }
    dimension = oldd - 1;	// one token for terminal label
    fclose(f);
    fprintf(stderr, "done: %d\n", dimension);

    return;
}


// main
// -----------------------------------------------------------------------------
int
main(int argc, char **argv) {

    FILE * f;

    // get command line options
    parse_options(argc, argv);
    
    // check for validity and get arity and dimension
    check();
    
    // initialization of environment
    lraamlib_init(arity, dimension, -1, -1, precision, randseed);
    
    // init of raam
    raam = lraam_alloc();
    if (raam == NULL) {
	fprintf(stderr, "ERROR: allocation memory for RAAM failed.\n\n");
	exit(1);
    }
    
    // randomize weights
    if (wbounds < 0) {
	lraam_randomize_weights(raam);
    } else {
	matrix_intrand(raam->W, -wbounds, wbounds);
    }

    // init of termset
    f = fopen(termset_filename, "r");
    if (f == NULL) {
	fprintf(stderr, "ERROR: termset file not opened.\n\n");
	print_usage();
	exit(1);
    }
    termset = set_load_terminals(f, dimension);
    if (termset == NULL) {
	fprintf(stderr, "ERROR: problems while loading termset.\n\n");
	exit(1);
    }
    fclose(f);

    // init of dataset
    f = fopen(dataset_filename, "r");
    if (f == NULL) {
	fprintf(stderr, "ERROR: dataset file not opened.\n\n");
	print_usage();
	exit(1);
    }
    dataset = set_load_structures(f, termset);
    if (dataset == NULL) {
	fprintf(stderr, "ERROR: problems while loading dataset.\n\n");
	exit(1);
    }
    fclose(f);

    // printing experiment data
    experiment_save(stdout, raam, dataset, termset, 0);
    
    // releasing memory
    lraam_free(raam);
    set_free(dataset);
    set_free(termset);

    return 0;
}
