/*!\file lraam_learning.c
 *
 * \author Michal Pokorny
 * \date 02/2006 - 03/2006
 *
 * \brief Utility for simulating learning process on given experiment
 *
 * Loads experiment from standard input (with matrix evolved to some 
 * extent). Then provides several learning iterations (specified by user) 
 * with given (also by user) learning rate. Writes on standard output 
 * experiment status after simulation.
 */

#include <getopt.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lraam.h"

#define MAX_PATH_LENGTH 128

#define DEFAULT_ETA 			0.01
#define DEFAULT_NUMBER_OF_ITERATIONS	3000

lraam_t * raam;
set_t * dataset;
set_t * termset;

// command line options defaults
// -----------------------------------------------------------------------------
value_t eta = DEFAULT_ETA;		/* switch --eta, 	-e	*/
int number_of_iterations = 
	DEFAULT_NUMBER_OF_ITERATIONS;	/* switch --iterations, -i	*/


// prints usage of this program
// -----------------------------------------------------------------------------
void
print_usage () {

    printf ("Usage:\n");
    printf ("  lraam_learning [OPTIONS]\n");
    printf ("\nOptions:\n");
    printf ("  -i <number_of_iterations>, --iterations=<number_of_iterations>\n");
    printf ("       Specify how many presentations of training set shold be done.\n");
    printf ("  -e <learning_rate>, --eta=<learning_rate>\n");
    printf ("       Set learning rate.\n");
    printf ("  -s, --settings\n");
    printf ("       Display default option values and exit.\n");
    printf ("  -h, --help\n");
    printf ("       Display this help and exit.\n");
    
    return;
}


// release memory allocated when loading experiment
// -----------------------------------------------------------------------------
void free_all() {

    lraam_free(raam);
    set_free(dataset);
    set_free(termset);

    return;
}


// prints default values of command line options
// -----------------------------------------------------------------------------
void
print_defaults () {

    printf ("Default values of command line options:\n");

    // eta parameter
    printf ("  <eta> = ");
    value_fprintf(stdout, eta);
    fprintf(stdout, "\n");
    // number of iterations
    printf ("  <number_of_iterations> = %d\n", number_of_iterations);
    
    return;
}


// parses command line options
// -----------------------------------------------------------------------------
void
parse_options(int argc, char **argv) {
    int c;
    static const struct option long_options[] = {
	{"eta", required_argument, NULL, 'e'},
	{"iterations", required_argument, NULL, 'i'},
	{"settings", no_argument, NULL, 's'},
	{"help", no_argument, NULL, 'h'},
	{0, 0, 0, 0}
    };

    static const char short_options[] = "e:i:sh";

    while(1) {
	c = getopt_long(argc, argv, short_options, long_options, NULL);
	if (c == -1) break;
	switch (c) {
	    case 'e': 
		sscanf(optarg, "%lf", &eta);
		break;
	    case 'i': 
		sscanf(optarg, "%d", &number_of_iterations);
		break;
	    case 's':
		print_defaults();
		exit(0);
		break;
	    case 'h':
	    case '?':
	    default:
		print_usage();
		exit(1);
	}
    }
    
    return;
}


// main
// -----------------------------------------------------------------------------
int
main(int argc, char **argv) {

    error_t err;
    
    // get command line options
    parse_options(argc, argv);

    // load experiment
    err = experiment_load(stdin, &raam, &dataset, &termset);
    if (err != ERR_OK) {
	fprintf(stderr, "ERROR %d: cannot load experiment.\n\n", err);
	exit(err);
    }

    // leraning process
    lraam_set_eta(raam, eta);
    err = lraam_set_presentation(raam, dataset, termset, number_of_iterations, NULL, 1, NULL);
    if (err != ERR_OK) {
	fprintf(stderr, "ERROR %d: learning process failed.\n\n", err);
	exit(err);
    }

    experiment_save(stdout, raam, dataset, termset, 0);

    // releasing memory
    free_all();

    return 0;
}
