/*!\file lraam_stat.c
 *
 * \author Michal Pokorny
 * \date 02/2006 - 03/2006
 *
 * \brief Utility for computing reconstruction error and for testing,
 *        which structures are encodable
 * 
 * In this version utility provides two main features.
 * Given experiment on standard input and probably set of structures for
 * testing (testset) via option value, utility can operate on this testset 
 * either by computing total reconstruction error or by testing each 
 * structure within testset, whether it is encodable by network, taking
 * into account theta reconstruction treshold variable set by user.\n
 * Structures which can network represent are printed as follows: for non-encodable
 * structure print blank line (i.e. "\n"), otherwise print the encodable
 * structure in single line. User can afterwards grep result to get all
 * encodable structures close together, and then, e.g., pipe result to
 * wc -l to get number of encodable structures.
 */

#include <getopt.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lraam.h"


#define MAX_PATH_LENGTH		128
#define DEFAULT_THETA 		0.7


lraam_t * raam;
set_t * dataset;
set_t * termset;
set_t * testset;


// command line options defaults
// -----------------------------------------------------------------------------
char testset_path[MAX_PATH_LENGTH] = "";	/* switch --testset, -t   */
value_t theta = DEFAULT_THETA;			/* switch --theta, -T	  */
int feature_error = 0;				/* switch --error, -e	  */
int feature_encodables = 0;			/* switch --encodable, -E */
int feature_representations = 0;		/* switch --representations, -r */


// prints usage of this program
// -----------------------------------------------------------------------------
void
print_usage () {

    printf ("Usage:\n");
    printf ("  lraam_output [OPTIONS]\n");
    printf ("\nOptions:\n");
    printf ("  -t <testset_path>, --testset=<testset_path>\n");
    printf ("       Set another structure set as testset. If this option is left,\n");
    printf ("       dataset saved in experiment status will be used instead.\n");
    printf ("  -T <theta>, --theta=<theta>\n");
    printf ("       Set the theta parameter - reconstruction treshold.\n");
    printf ("  -e, --error\n");
    printf ("       Output total reconstruction error over structures in testset.\n");
    printf ("  -E, --encodable\n");
    printf ("       Print statistics, which structures from dataset can be encoded\n");
    printf ("       successfully decode by raam with respect to theta parameter set\n");
    printf ("       by user and current weight matrix.\n");
    printf ("  -r, --representations\n");
    printf ("       Print encodables structures together with their representations.\n");
    printf ("  -s, --settings\n");
    printf ("       Display default option values and exit.\n");
    printf ("  -h, --help\n");
    printf ("       Display this help and exit.\n");
    
    return;
}


// release memory allocated when loading experiment
// -----------------------------------------------------------------------------
void free_all() {

    lraam_free(raam);
    if (testset != dataset) set_free(testset);
    set_free(dataset);
    set_free(termset);

    return;
}


// prints default values of command line options
// -----------------------------------------------------------------------------
void
print_defaults () {

    printf ("Default values of command line options:\n");

    // theta
    printf ("  <theta> = ");
    value_fprintf(stdout, theta);
    fprintf(stdout, "\n");
    
    // testset
    printf ("  <testset> = ");
    if (!strcmp(testset_path, "")) {
	printf("dataset from experiment");
    } else {
	printf("%s", testset_path);
    }
    fprintf(stdout, "\n");
    
    // features
    printf ("  <error> = %s\n", feature_error ? "yes" : "no");
    printf ("  <encodable> = %s\n", feature_encodables ? "yes" : "no");
    printf ("  <representations> = %s\n", feature_representations ? "yes" : "no");
    
    return;
}


// parses command line options
// -----------------------------------------------------------------------------
void
parse_options(int argc, char **argv) {
    int c;
    
    static const struct option long_options[] = {
	{"testset", required_argument, NULL, 't'},
	{"theta", required_argument, NULL, 'T'},
	{"error", no_argument, NULL, 'e'},
	{"encodable", no_argument, NULL, 'E'},
	{"representations", no_argument, NULL, 'r'},
	{"settings", no_argument, NULL, 's'},
	{"help", no_argument, NULL, 'h'},
	{0, 0, 0, 0}
    };

    static const char short_options[] = "t:T:eErsh";

    while(1) {
	c = getopt_long(argc, argv, short_options, long_options, NULL);
	if (c == -1) break;
	switch (c) {
	    case 't': 
		sscanf(optarg, "%s ", testset_path);
		break;
	    case 'T': 
		sscanf(optarg, "%lf", &theta);
		break;
	    case 'e': 
		feature_error = 1;
		break;
	    case 'E': 
		feature_encodables = 1;
		break;
	    case 'r': 
		feature_representations = 1;
		break;
	    case 's':
		print_defaults();
		exit(0);
		break;
	    case 'h':
	    case '?':
	    default:
		print_usage();
		exit(1);
	}
    }
    
    return;
}


// main
// -----------------------------------------------------------------------------
int
main(int argc, char **argv) {

    error_t err;
    double recerr;
    bool_t * encodables;
    FILE * f;
    int i;

    // get command line options and load experiment if needed
    parse_options(argc, argv);

    // loading experiment
    err = experiment_load(stdin, &raam, &dataset, &termset);
    if (err != ERR_OK) {
	fprintf(stderr, "ERROR %d: cannot load experiment.\n\n", err);
	exit(err);
    }
    
    // set theta reconstruction treshold parameter
    lraam_set_theta(raam, theta);

    // load testset, if specified
    if (strcmp(testset_path, "")) {
	f = fopen(testset_path, "r");
	if (f == NULL) {
	    fprintf(stderr, "ERROR: testset not found: %s.\n\n", testset_path);
	    print_usage();
	    exit(1);
	}
	testset = set_load_structures(f, termset);
	if (testset == NULL) {
	    fprintf(stderr, "ERROR: cannot load testset.\n\n");
	    exit(1);
	}
	fclose(f);
    } else {
	testset = dataset;
    }

    // compute reconstruction error
    if (feature_error) {
	recerr = -1.0;
	lraam_set_presentation(raam, testset, termset, 1, &recerr, 0, NULL);
	fprintf(stdout, "%.10lf\t", recerr);
    }
    
    // print encodable structures
    if (feature_encodables) {
        // memory for array encodables[]
	encodables = (bool_t *)calloc(testset->no_of_objects, sizeof(bool_t));
	if (encodables == NULL) {
	    fprintf(stderr, "ERROR: encodables[]: not enought memory.\n\n");
	    exit(err);
	}
	// testing which structures can network represent (are encodable)
	lraam_set_presentation(raam, testset, termset, 1, NULL, 0, encodables);
	// output
	for (i=0; i<testset->no_of_objects; i++) {
	    if (encodables[i]) {
		struct_fprintf(stdout, testset->objects[i], termset, feature_representations);
	    } else {
		fprintf(stdout, "\n");
	    }
	}
	free(encodables);
    }

    // releasing memory
    free_all();

    return 0;
}
